// stack.h - implementation of a template Stack class
//           This version uses friends rather than accessor functions

// Note: cannot be broken off to a stack.cpp implementation file
//       due to longstanding bug(s) in the Visual C++ compiler

#ifndef STACK_H
#define STACK_H

#include <iostream>
#include <string>

using namespace std;

// Forward declaration is necessary because Stack is used in StackFrame
template <class T>
class Stack;

template <class T>
class StackFrame {
  private:
    T data;
    StackFrame *link;

  public:
    StackFrame(T data, StackFrame* link = NULL) : data(data), link(link) {}

    friend class Stack<T>;
    friend ostream& operator <<(ostream& outs, Stack<T>& aStack);
};

template <class T>
class Stack {
  private:
    StackFrame<T>* top;

  public:    
    Stack() {
	top = NULL;
    }

    Stack(const Stack<T>& aStack) {
	if (aStack.top == NULL)
	    top = NULL;
	else {
	    StackFrame<T>* ptr = aStack.top;
	    StackFrame<T>* end = new StackFrame<T>(ptr->data);
	    top = end;

	    while (ptr = ptr->link, ptr != NULL) {
		end->link = new StackFrame<T>(ptr->data);
		end = end->link;
	    }
	}
    }

    ~Stack() {
	while (! empty())
	    pop();
    }

    void push(T data) {
	top = new StackFrame<T>(data, top);
    }

    T pop() {
	if (empty()) {
	    cerr << "Error: attempted pop of empty stack" << endl;
	    exit(1);
	}
	T data = top->data;
	StackFrame<T>* ptr = top;
	top = top->link;
	delete ptr;
	return data;
    }

    bool empty() const {
	return (top == NULL);
    }

    friend ostream& operator <<(ostream& outs, Stack& aStack) {
        outs << "(";
	for (StackFrame<T>* ptr = aStack.top; ptr != NULL; ptr = ptr->link)
	    outs << " " << ptr->data;
        outs << " )";
        return outs;
    }

};

#endif
